/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "conf.h"
#include <sstream>
#include <algorithm>
#include <set>

#define ERR_READ    0
#define ERR_WRITE   1
#define ERR_REPLACE 2

//-----------------------------------------------------------------------------
// Check if given assignment has no values

bool EmptyAssgn(const vector<string> &assgn){

    return assgn.size() < 2 ||
        all_of(assgn.begin() + 1, assgn.end(),
            [](const string& param){return param.empty();});
    }

//-----------------------------------------------------------------------------
// Form a basic error message for a specified type of error

const string ErrMsg(const int err_type, const string &sec_name,
        const string &var_name){

    const string param = sec_name + ":" + var_name + " parameter";

    switch(err_type){

        case ERR_READ:
            return "Can't read " + param + " of the DOM";

        case ERR_WRITE:
            return "Can't write value of a " + param + " "
                "of the configuration file";
        }

    return "";
    }

const string ErrMsg(const int err_type, const int n){

    switch(err_type){

        case ERR_REPLACE:
            return "Can't replace line " + to_string(n) + " "
                "of the configuration file";
        }

    return "";
    }

//-----------------------------------------------------------------------------
// Get vector of values of an assignment for a specified variable name

vector<string> TConfSec::VarAssgn(const string &var_name) const{

    for(auto a = assgn.rbegin(); a != assgn.rend(); a++){
        if(a->size() > 1 && a->front() == var_name){
            return vector<string>(a->begin() + 1, a->end());
            }
        }

    return {};
    }

//-----------------------------------------------------------------------------
// Get first value of a specified variable

string TConfSec::VarSingleAssgn(const string &var_name) const{

    for(auto a = assgn.rbegin(); a != assgn.rend(); a++){
        if(a->size() > 1 && a->front() == var_name){
            return a->at(1);
            }
        }

    return "";
    }

//-----------------------------------------------------------------------------
// Change assignment of a variable that has a name that stored in a first
// element of a given vector

#define ERR_MSG "Can't change an assignment"

bool TConfSec::ChangeAssgn(const vector<string> &new_assgn, string &err){

    if(new_assgn.size() < 2){
        err = ERR_MSG " - an assignment must contain at least two elements";
        return false;
        }

    for(auto a = (assgn.end() - 1); a != (assgn.begin() - 1); a--){
        if(a->front() == new_assgn.front()){
            if(EmptyAssgn(new_assgn)){
                assgn.erase(a);
                }
            else{
                *a = new_assgn;
                }
            return true;
            }
        }

    err = ERR_MSG " - there is no '" + new_assgn.front() + "' item";

    return false;
    }

//-----------------------------------------------------------------------------
// Rewrite i-th assignment

bool TConfSec::ChangeAssgn(const vector<string> &new_assgn, const int i,
        string &err){

    if((int)new_assgn.size() < 2){
        err = ERR_MSG " - an assignment must contain at least two elements";
        return false;
        }

    if((int)assgn.size() <= i){
        err = ERR_MSG " - there are less than a specified number "
            "of assignments (" + to_string(i) + ")";
        return false;
        }

    if(EmptyAssgn(new_assgn)){
        assgn.erase(assgn.begin() + i);
        }
    else{
        assgn[i] = new_assgn;
        }

    return true;
    }

#undef ERR_MSG

//-----------------------------------------------------------------------------
// Display DOM

void TConfDOM::Print(void) const{

    int sec_count = 0;

    for(const auto &s : sec){

        cout << "Section" << sec_count++ << " - " << s.Name() << ":" << endl;

        for(const auto &a : s.Assgn()){

            if(a.empty()){
                continue;
                }

            string a_str;

            for(const auto &w : a){
                a_str += w + (a_str.empty() ? "={" : ",");
                }

            if(a.size() > 1){
                a_str.pop_back();
                }
            a_str += '}';

            cout << "\t" << a_str << endl;
            }
        }
    }

//-----------------------------------------------------------------------------
// Get index of a section that has a specified name

int TConfDOM::FindSec(const string &sec_name) const{

    for(size_t i = 0; i < sec.size(); i++){
        if(sec_name == sec[i].Name()){
            return i;
            }
        }

    // There is no section with a specified name
    return -1;
    }

//-----------------------------------------------------------------------------
// Get index of an assignment of a variable that exists in a specified section
// and that has a specified name

int TConfDOM::FindAssgn(const int sec_num, const string &var_name) const{

    if(sec_num >= (int)sec.size()){
        return -1;
        }

    int counter = 0;

    for(const auto &a : sec.at(sec_num).Assgn()){
        if(!a.empty() && var_name == a.front()){
            return counter;
            }
        counter++;
        }

    // There is no assignment with the given params
    return -1;
    }

//-----------------------------------------------------------------------------
// Determine section number and assignment number for a specified section name
// and a specified variable name

bool TConfDOM::Find(const string &sec_name, const string &var_name,
        int &sec_num, int &assgn_num, string &err) const{

    sec_num = FindSec(sec_name);
    assgn_num = -1;

    if(sec_num >= 0){
        assgn_num = FindAssgn(sec_num, var_name);
        if(assgn_num >= 0){
            return true;
            }
        else{
            err = "Can't find item '" + var_name + "' "
                "in section '" + sec_name + "'";
            }
        }
    else{
        err = "Can't find section '" + sec_name + "'";
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Insert given section to DOM

bool TConfDOM::InsertSec(const int index, const TConfSec &s){

    if(index < (int)sec.size()){
        sec.insert(sec.begin() + index, s);
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Write section that has a specified index to a referenced object

bool TConfDOM::GetSec(const int i, TConfSec &s) const{

    if(i >= 0 && i < (int)sec.size()){
        s = sec.at(i);
        return true;
        }

    return false;
    }

//----------------------------------------------------------------------------
// Write section that has a specified name to a referenced object

bool TConfDOM::GetSec(const string &sec_name, TConfSec &sec_in) const{

    for(auto &s : sec){
        if(s.Name() == sec_name){
            sec_in = s;
            return true;
            }
        }

    return false;
    }

//----------------------------------------------------------------------------
// Write last/top DOM section to a referenced object

bool TConfDOM::GetLastSec(TConfSec &s) const{

    if(!sec.empty()){
        s = sec.back();
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Add given assignment to a specified section, create the section if
// the 'force' flag is passed

bool TConfDOM::PushAssgnToSec(const vector<string> &a, const string &sec_name,
        string &err, const bool force){

    const bool empty_assgn = EmptyAssgn(a);
    const int n = FindSec(sec_name);

    if(n >= 0){
        if(!empty_assgn){
            sec[n].AddAssgn(a);
            }
        return true;
        }
    else if(force){
        TConfSec new_sec;
        new_sec.WriteName(sec_name);
        if(!empty_assgn){
            new_sec.AddAssgn(a);
            }
        sec.push_back(new_sec);
        return true;
        }

    err = "Can't push assignment to a specified section: "
        "There is no '" + sec_name + "' section in the DOM";

    return false;
    }

//-----------------------------------------------------------------------------
// Drop section that has a given index

bool TConfDOM::DropSec(const int index){

    if(index >= 0 && index < (int)sec.size()){
        sec.erase(sec.begin() + index);
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Read conf file to memory

bool TConf::ReadFile(const string &path){

    string conf_file_path;

    if(path.empty()){
        if(file_path.empty()){
            ErrPush("There is no configuration file path");
            return false;
            }
        conf_file_path = file_path;
        }
    else{
        conf_file_path = path;
        }

     if(!file.ReadFileToMem(conf_file_path)){
        ErrPush("Can't read configuration file to memory");
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Parse conf file and form a DOM

bool TConf::Parse(void){

    // Clear current DOM
    dom.Clear();

    // Put conf file data string into string stream
    istringstream iss(file.data);

    string line;
    TConfSec sec;
    vector<string> assgn;

    // For all lines of the string stream
    while(getline(iss,line)){
        if(!ParseNextLine(line, sec, assgn)){
            continue;
            }
        // If current line is section title
        if(!sec.Name().empty()){
            // Adding a section into the DOM
            dom.AddSec(sec);
            continue;
            }
        // If there are no sections in the DOM yet
        if(dom.Sections().empty()){
            // Assignment can exist only inside a section
            continue;
            }
        // If current line is an assignment line
        if(!assgn.empty()){
            dom.AddAssgn(assgn);
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Determine indexes of opening and closing quotation marks

bool TConf::FindQtnMarks(const string &line, size_t &start, size_t &stop) const{

    // Search for double quotation marks
    const size_t open_double_q_mark_pos = UnescapedQtnMarkPos(line, 0, true);
    const size_t close_double_q_mark_pos =
        open_double_q_mark_pos != string::npos ?
            UnescapedQtnMarkPos(line, open_double_q_mark_pos + 1, true) :
            string::npos;

    // Search for single quotation marks
    const size_t open_sinqle_q_mark_pos = UnescapedQtnMarkPos(line, 0, false);
    const size_t close_sinqle_q_mark_pos =
        open_sinqle_q_mark_pos != string::npos ?
            UnescapedQtnMarkPos(line, open_sinqle_q_mark_pos + 1, false) :
            string::npos;

    // Detect pair of quotation marks (opening and closing)
    bool double_q_mark = open_double_q_mark_pos != string::npos &&
        close_double_q_mark_pos != string::npos;
    bool sinqle_q_mark = open_sinqle_q_mark_pos != string::npos &&
        close_sinqle_q_mark_pos != string::npos;

    // If there are no quotation marks
    if(!double_q_mark && !sinqle_q_mark){
        start = stop = string::npos;
        return false;
        }

    // Determine priority if both pairs (double and single) exist
    if(double_q_mark && sinqle_q_mark){
        if(open_sinqle_q_mark_pos < open_double_q_mark_pos){
            double_q_mark = false;
            }
        }

    // Determine opening and closing indexes
    if(double_q_mark){
        start = open_double_q_mark_pos;
        stop = close_double_q_mark_pos;
        }
    else{
        start = open_sinqle_q_mark_pos;
        stop = close_sinqle_q_mark_pos;
        }

    // There should be no comma before opening quote
    // because it can be closing quote for the previous value
    size_t com_pos = line.find(',');
    if(com_pos != string::npos && com_pos < start){
        start = stop = string::npos;
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Assign next quote from a given line to the 'word' var and determine indexes
// of opening and closing quotation marks

bool TConf::GetQuote(const string &line, string &word, size_t &start,
        size_t &stop) const{

    if(FindQtnMarks(line, start, stop)){
        word.assign(line.substr(start + 1, stop - start - 1));
        return true;
        }

    word.clear();

    return false;
    }

//-----------------------------------------------------------------------------
// Parse given line of configuration file

bool TConf::ParseNextLine(const string &line, TConfSec &sec,
        vector<string> &assgn) const{

    const string filtered_line = CutComment(line);

    sec.Clear();

    // If this line is a section header
    const string sec_name = NextSec(filtered_line);
    if(!sec_name.empty()){
        sec.WriteName(sec_name);
        return true;
        }

    // If this line is an assignment
    const vector<string> new_assgn = NextAssgn(filtered_line);
    if(new_assgn.empty()){
        return false;
        }

    assgn = new_assgn;

    return true;
    }

//-----------------------------------------------------------------------------
// Get name of a next section

string TConf::NextSec(const string &line) const{

    string sec_name;

    const int open_bracket_pos = line.find('[');
    if(open_bracket_pos < 0){
        return "";
        }

    // Only white spaces allowed before an opening bracket
    for(int i = 0; i < open_bracket_pos; i++){
        if(!isspace(line.at(i))){
            return "";
            }
        }

    const int close_bracket_pos = line.find(']');
    if((close_bracket_pos - open_bracket_pos) <= 0){
        return "";
        }

    // For all chars in square brackets
    for(int i = (open_bracket_pos + 1); i < close_bracket_pos; i++){
        // If allowed symbol
        if(!isspace(line.at(i))){
            sec_name += line.at(i);
            }
        else if(!sec_name.empty()){
            break;
            }
        }

    return sec_name;
    }

//-----------------------------------------------------------------------------
// Get vector of names of all sections of a conf

vector<string> TConf::AllSecNames(void) const{

    vector<string> sec_names;
    istringstream iss(file.data);
    string line;

    while(getline(iss,line)){
        const string sec_name = NextSec(line);
        if(!sec_name.empty()){
            sec_names.push_back(sec_name);
            }
        }

    return sec_names;
    }

//-----------------------------------------------------------------------------
// Get vector of names of all variables of a specified (or last) section

vector<string> TConf::AllVarNamesOfSec(const string &sec_name){

    TConfSec s;

    if(!sec_name.empty() && !dom.GetSec(sec_name, s)){
        ErrPush("There is no '" + sec_name + "' section");
        return {};
        }

    if(sec_name.empty() && !dom.GetLastSec(s)){
        ErrPush("There is not a single section in the configuration file");
        return {};
        }

    vector<string> var_names_vec;
    set<string> var_names_set;

    for(const auto &a : s.Assgn()){
        if(a.size() < 2){
            continue;
            }
        const auto insert_ret = var_names_set.insert(a.front());
        // Values of the formed vector should be unique
        // - later assign overlaps the earlier assign
        if(insert_ret.second == false){
            var_names_vec.erase(
                find(var_names_vec.begin(), var_names_vec.end(), a.front())
                );
            }
        var_names_vec.push_back(a.front());
        }

    return var_names_vec;
    }

//-----------------------------------------------------------------------------
// Search for and get next assignment from a given line

vector<string> TConf::NextAssgn(string line) const{

    // There should be at least 1 symbol before and after assign symbol
    if(line.size() < 3){
        return {};
        }

    vector<string> assgn;
    string word;

    size_t pos = line.find('=');

    if(pos == string::npos || pos > (line.size() - 2)){
        return {};
        }

    // Finding var name
    if(GetWordFromLine(line, word, 0, pos) != string::npos){
        // Adding var name into assgn mass
        assgn.push_back(word);
        }
    else{
        return {};
        }

    // Handle right part of assignment
    size_t start = string::npos;
    size_t stop  = string::npos;
    do{
        // Cut processed part of line
        line.assign(line.substr(pos + 1));
        // Finding next value
        if(GetQuote(line, word, start, stop) ||
                GetWordFromLine(line, word) != string::npos){
            // Adding next value into assgn mass
            assgn.push_back(Unesc(word));
            // Searching for a delimiter and making a shift for a later cut
            // of processed part of a line
            pos = (stop != string::npos) ?
                line.find(',', stop) :
                line.find(',');
            }
        }
    while(!word.empty() && pos != string::npos);

    return assgn;
    }

//-----------------------------------------------------------------------------
// Form an assignment string

string TConf::AssgnStr(const vector<string> &assgn) const{

    string str;

    for(auto el = assgn.begin(); el != assgn.end(); el++){
        if(el > (assgn.begin() + 1)){
            str += ", ";
            }
        else if(el == (assgn.begin() + 1)){
            str += " = ";
            }
        str += *el;
        }

    return str;
    }

//-----------------------------------------------------------------------------
// Read first value of a specified variable from a given section and write
// this value to the 'param' ref

bool TConf::ReadDOMParam(string &param, const TConfSec &s, const string &var_name){

    const vector<string> assgn = s.VarAssgn(var_name);

    if(assgn.empty()){
        param.clear();
        ErrPush(ErrMsg(ERR_READ, s.Name(), var_name));
        return false;
        }

    param = assgn.front();

    return true;
    }

//-----------------------------------------------------------------------------
// Read first value of a specified variable from a section that has a specified
// name and write this value to the 'param' var

bool TConf::ReadDOMParam(string &param, const string &sec_name,
        const string &var_name){

    int sec_num, assgn_num;

    // Determine a section number and an assignment number
    // for a given section name and a variable name
    if(dom.Find(sec_name, var_name, sec_num, assgn_num, ErrNew())){
        const auto &sec = dom.at(sec_num);
        const auto &assgn = sec.Assgn().at(assgn_num);
        if(assgn.size() > 1){
            param.assign(assgn.at(1));
            return true;
            }
        }

    param.clear();

    ErrPush(ErrMsg(ERR_READ, sec_name, var_name));

    return false;
    }

//-----------------------------------------------------------------------------
// Read vector of values of a specified variable of a specified section and
// write these values to 'vparam'

bool TConf::ReadVParams(vector<string> &vparam, const string &sec_name,
        const string &var_name){

    int sec_num, assgn_num;

    vparam.clear();

    if(dom.Find(sec_name, var_name, sec_num, assgn_num, ErrNew())){
        const auto &sec = dom.at(sec_num);
        const auto &assgn = sec.Assgn().at(assgn_num);
        vparam.insert(vparam.end(), assgn.begin() + 1, assgn.end());
        return true;
        }

    ErrPush(ErrMsg(ERR_READ, sec_name, var_name));

    return false;
    }

//-----------------------------------------------------------------------------
// Write value of a specified variable of a specified section

bool TConf::ChangeDOMParam(const string &val, const string &sec_name,
        const string &var_name, const bool force){

    bool success = false;
    int sec_num, assgn_num;
    const vector<string> new_assgn = {var_name, val};

    // Determining a section number and an assignment number
    if(dom.Find(sec_name, var_name, sec_num, assgn_num, ErrNew())){
        auto &s = dom[sec_num];
        auto &assgn = s.Assgn().at(assgn_num);
        if(assgn.size() > 1){
            success = s.ChangeAssgn(new_assgn, assgn_num, ErrNew());
            }
        }
    else if(force){
        success = dom.PushAssgnToSec(new_assgn, sec_name, ErrNew(), force);
        }

    if(!success){
        ErrPush("Can't change " + sec_name + ":" + var_name + " parameter "
            "of the DOM");
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Rewrite configuration file completly according to DOM

bool TConf::RewriteAccordingToDOM(const string &header){

    string &data = file.data;

    if(!header.empty()){
        data = header + "\n";
        }

    for(const auto &sec : dom.Sections()){
        if(!data.empty()){
            data += '\n';
            }
        data += "[" + sec.Name() + "]\n";
        for(const auto &assgn : sec.Assgn()){
            data += AssgnStr(assgn) + "\n";
            }
        }

    if(file.WriteBackToFile()){
        return true;
        }

    ErrPush("Can't transfer the DOM of the configuration file back to the file");

    return false;
    }

//-----------------------------------------------------------------------------
// Write new assignent back into conf file data string

bool TConf::WriteParam(const string &sec_name, const string &var_name,
        const bool write_to_file){

    int sec_num;
    int assgn_num;
    bool success = false;

    // Determining a section number and an assignment number
    if(dom.Find(sec_name, var_name, sec_num, assgn_num, ErrNew())){

        const auto &s = dom.at(sec_num);
        const auto &assgn = s.Assgn().at(assgn_num);

        if(assgn.size() < 2){
            ErrPush(ErrMsg(ERR_WRITE, sec_name, var_name) +
                ": The parameter has no assigned value");
            return false;
            }

        // Putting a conf file data string into string stream
        istringstream iss(file.data);
        string current_line;

        // Finding a specified section in a conf-file data string
        TConfSec current_sec;
        vector<string> current_assgn;
        int lnum = -1;
        int blank_count = 0;
        int blank_count0 = 0;
        bool sec_found = false;
        while(getline(iss,current_line)){
            lnum++;
            Trim(current_line);
            if(current_line.empty()){
                if(sec_found){
                    blank_count++;
                    }
                continue;
                }
            if(blank_count == blank_count0){
                blank_count = 0;
                }
            ParseNextLine(current_line, current_sec, current_assgn);
            // If current line is a section title that is matches
            // a specified section name
            if(!sec_found && !current_sec.Name().empty() &&
                    current_sec.Name() == sec_name){
                sec_found = true;
                }
            // Section is found but assignment for a specified var name is not
            else if(sec_found && !current_sec.Name().empty()){
                success = AddNewAssgn(lnum - blank_count, assgn);
                break;
                }
            // Section is found as well as assignment for a specified var name
            else if(sec_found && !current_assgn.empty() &&
                    current_assgn.front() == var_name){
                success = ReplaceLine(lnum, assgn);
                break;
                }
            blank_count0 = blank_count;
            }

        // Section is found as well as EOF
        if(!success && sec_found){
            success = AddNewAssgn(lnum + 1 - blank_count, assgn);
            }
        }

    if(!success){
        ErrPush(ErrMsg(ERR_WRITE, sec_name, var_name));
        }

    // Rewriting the conf file
    return success && (!write_to_file || file.WriteBackToFile());
    }

//-----------------------------------------------------------------------------
// Replace given line of a conf file data string with a string representation
// of a given assignment

bool TConf::ReplaceLine(const int n, const vector<string> &new_assgn){

    // Searching for a starting index of a n-th line
    const int start = file.NLineIndex(n);
    if(start < 0){
        ErrPush(
            ErrMsg(ERR_REPLACE, n) + ": Can't get line " + to_string(n));
        return false;
        }

    // Determining a length of a n-th line
    const int len = file.NLineIndex(1, start) - start - 1;
    if(len < 0){
        ErrPush(
            ErrMsg(ERR_REPLACE, n) + ": Can't determine a length of the line");
        return false;
        }

    // Putting a new assignment into output string stream
    ostringstream oss;
    for(size_t i = 0; i < new_assgn.size(); i++){
        oss << new_assgn[i];
        if(i == 0){
            oss << " = ";
            }
        else if(i < (new_assgn.size() - 1)){
            oss << " , ";
            }
        }

    // Writing a string stream content into conf file data string
    file.data.replace(start, len, oss.str());

    return true;
    }

//-----------------------------------------------------------------------------
// Add new line to a conf file data string

bool TConf::AddNewLine(const int next_line_n, const string &text){

    const int pos = file.NLineIndex(next_line_n);

    if(pos < 0){
        ErrPush("Can't add new line to the configuration file");
        return false;
        }

    file.data.insert(pos, text);

    return true;
    }

//-----------------------------------------------------------------------------
// Prepare an empty string for a new error message

string& TConf::ErrNew(void){

    if(err.empty() || !err.back().empty()){
        err.push_back("");
        }

    return err.back();
    }

//-----------------------------------------------------------------------------
// Push new error message to the vector while keeping in mind a possibility of
// presence of a specially crafted empty string at the back

void TConf::ErrPush(const string &msg){

    if(!err.empty() && err.back().empty()){
        err.back() = msg;
        }
    else{
        err.push_back(msg);
        }
    }

//-----------------------------------------------------------------------------
// Get a string of joined error messages starting from a last one

const string TConf::Err(void) const{

    string err_str;

    for(auto msg = err.rbegin(); msg != err.rend(); msg++){
        if(!msg->empty()){
            if(!err_str.empty()){
                err_str += "; ";
                }
            err_str += *msg;
            }
        }

    return err_str;
    }

//-----------------------------------------------------------------------------
