/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef lcontrolpp_confH
#define lcontrolpp_confH

#include "libcontrol++.h"
#include "text.h"
#include "filesystem.h"

#define CONF_MEM_STEP  4    // Step of memory allocation for conf DOM

//-----------------------------------------------------------------------------

// Check if given assignment has no values
bool EmptyAssgn(const vector<string> &assgn);

//-----------------------------------------------------------------------------
// Section of configuration file

class TConfSec{

    private:
        string name;   // Section name
        // Assignments of a section
        // - 1st index - assignment number
        // - 0th element of a 2nd index - variable name of a current assgn
        // - {1, 2, ..} elements of a 2nd index - values of a current assgn
        vector<vector<string> > assgn;

    public:
        // Data access
        const string& Name(void) const{return name;}
        const vector<vector<string> >& Assgn(void) const{return assgn;}
        vector<string> VarAssgn(const string &var_name) const;
        string VarSingleAssgn(const string &var_name) const;
        // Modifiers
        void WriteName(const string &new_name=""){
            name = new_name;
            }
        void WriteAssgn(const vector<vector<string> > &new_assgn){
            assgn = new_assgn;
            }
        void AddAssgn(const vector<string> &new_assgn){
            assgn.push_back(new_assgn);
            }
        bool ChangeAssgn(const vector<string> &new_assgn, string &err);
        bool ChangeAssgn(const vector<string> &new_assgn, const int i, string &err);

        void Clear(void){
            name.clear();
            assgn.clear();
            }
        TConfSec(void){}
        TConfSec(const string &sec_name){
            name = sec_name;
            }
        ~TConfSec(void){}
    };

//-----------------------------------------------------------------------------
// Configuration file DOM

class TConfDOM{

    private:
        vector<TConfSec> sec;   // Sections of the configuration file

    public:
        // Data access
        const vector<TConfSec>& Sections(void) const{return sec;}
        TConfSec& operator[](int i){return sec[i];}
        const TConfSec& at(int i) const{return sec.at(i);}
        // Display DOM
        void Print(void) const;
        // Get index of a section that has a specified name
        int FindSec(const string &sec_name) const;
        // Get index of an assignment of a variable that exists in a specified
        // section and that has a specified name
        int FindAssgn(const int sec_num, const string &var_name) const;
        // Determine section number and assignment number for a specified
        // section name and a specified variable name
        bool Find(const string &sec_name, const string &var_name, int &sec_num,
                int &assgn_num, string &err) const;
        // Add section to DOM
        void AddSec(const TConfSec &s){sec.push_back(s);}
        // Insert given section to DOM
        bool InsertSec(const int index, const TConfSec &s);
        // Add assignment to a last section
        void AddAssgn(vector<string> &a){sec[sec.size() - 1].AddAssgn(a);}
        // Add given assignment to a specified section, create the section if
        // the 'force' flag is passed
        bool PushAssgnToSec(const vector<string> &a, const string &sec_name,
                string& err, const bool force=false);
        // Write section that has a specified index to a referenced object
        bool GetSec(const int i, TConfSec &s) const;
        // Write section that has a specified name to a referenced object
        bool GetSec(const string &sec_name, TConfSec &s) const;
        // Write last/top DOM section to a referenced object
        bool GetLastSec(TConfSec &s) const;
        // Drop section that has a given index
        bool DropSec(const int index);
        // Drop last section
        bool DropLastSec(void){return DropSec(sec.size() - 1);}

        void Clear(void){sec.clear();}
        TConfDOM(void){}
        ~TConfDOM(void){}
    };

//-----------------------------------------------------------------------------
// Configuration file handler

class TConf{

    private:
        string file_path;   // Full path to configuration file
        TFileInMem file;    // Configuration file in memory
        TConfDOM dom;       // Configuration DOM
        vector<string> err; // Error messages

        // Determine indexes of opening and closing quotation marks
        bool FindQtnMarks(const string &line, size_t &start, size_t &stop) const;
        // Assign next quote from a given line to the 'word' var and determine
        // indexes of opening and closing quotation marks
        bool GetQuote(const string &line, string &word, size_t &start,
                size_t &stop) const;
        // Parse given line of configuration file
        bool ParseNextLine(const string &line, TConfSec &sec,
                vector<string> &assgn) const;
        // Get name of a next section
        string NextSec(const string &line) const;
        // Search for and get next assignment from a given line
        vector<string> NextAssgn(string line) const;
        // Form an assignment string
        string AssgnStr(const vector<string> &assgn) const;
        // Add new line to a conf file data string
        bool AddNewLine(const int next_line_n, const string &text);
        // Add string representation of a given assignment to a conf data string
        bool AddNewAssgn(const int next_line_n, const vector<string> &new_assgn){
            return AddNewLine(next_line_n, AssgnStr(new_assgn) + "\n");
            }
        // Replace given line of a conf data string with a string
        // representation of a given assignment
        bool ReplaceLine(const int n, const vector<string> &new_assgn);

        string& ErrNew(void);
        void ErrPush(const string &msg);

    public:
        // Display conf file data
        void PrintFile(void) const{
            cout << "Content of the configuration file:" << endl << file.data;
            }
        // Display configuration DOM
        void PrintDOM(void) const{dom.Print();}
        // Get full path of the configuration file
        const string& FilePath(void){return file_path;}
        // Write full path of the configuration file
        void WriteFilePath(const string &path=""){file_path=path;}
        // Read conf file to memory
        bool ReadFile(const string &path="");
        // Parse conf file and form a DOM
        bool Parse(void);
        // Get vector of all the sections of the conf data
        const vector<TConfSec>& AllSec(void) const{return dom.Sections();}
        // Get vector of names of all sections of a conf
        vector<string> AllSecNames(void) const;
        // Get vector of names of all variables of a specified (or last) section
        vector<string> AllVarNamesOfSec(const string &sec_name="");
        // Read first value of a specified variable from a given section
        // and write this value to the 'param' var
        bool ReadDOMParam(string &param, const TConfSec &sec, const string &var);
        // Read first value of a specified variable from a section that has a
        // specified name and write this value to the 'param' var
        bool ReadDOMParam(string &param, const string &sec, const string &var);
        // Read vector of values of a specified variable of a specified section
        // and write these values to 'vparam'
        bool ReadVParams(vector<string> &vparam, const string &sec_name,
                const string &var_name);
        // Write value of a specified variable of a specified section
        bool ChangeDOMParam(const string &val, const string &sec,
                const string &var, const bool force=false);
        // Add section to a DOM of the conf file
        void AddDOMSec(const TConfSec &s){dom.AddSec(s);}
        // Insert section to a DOM of the conf file
        bool InsertDOMSec(const int index, const TConfSec &sec){
            return dom.InsertSec(index, sec);
            }
        // Clear DOM of the conf file
        void CleanDOM(void){dom.Clear();}
        // Drop section with a given index from the DOM of the conf
        bool DropDOMSec(const int index){return dom.DropSec(index);}
        // Drop last section from the DOM of the conf
        bool DropLastDOMSec(void){return dom.DropLastSec();}
        // Get number of DOM sections
        int DOMSize(void){return dom.Sections().size();}
        // Rewrite configuration file completly according to DOM
        bool RewriteAccordingToDOM(const string &header="");
        // Write new assignent back into conf file data string
        bool WriteParam(const string &sec, const string &var,
                const bool write_to_file=true);
        // Check whether section with the given name exists
        bool CheckSec(const string &sec_name) const{
            return (dom.FindSec(sec_name) >= 0);
            }
        // Get DOM section with given index
        bool GetDOMSec(const int i, TConfSec &s) const{
            return dom.GetSec(i, s);
            }
        // Get DOM section with given name
        bool GetDOMSec(const string &sec_name, TConfSec &s) const{
            return dom.GetSec(sec_name, s);
            }
        // Get last DOM section
        bool GetLastDOMSec(TConfSec &s) const{return dom.GetLastSec(s);}

        const string Err(void) const;
        void ErrClear(void){err.clear();}

        TConf(void){}
        ~TConf(void){}
    };

//-----------------------------------------------------------------------------

#endif
