/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem.h"
#include <fstream>
#include <pwd.h>
#include <sys/stat.h>
#include <unistd.h>
#include <map>
#include <algorithm>

//-----------------------------------------------------------------------------
// Check file existence

bool CheckFile(const string &path){

    if(path.empty()){
        return false;
        }

    ifstream file(path);

    return file.good();
    }

//-----------------------------------------------------------------------------
// Check file extension

bool CheckExtension(const string &filename, const string &ext){

    if(filename.empty() || ext.empty()){
        return false;
        }

    const string ext_compl = (ext.front() != '.') ? ("." + ext) : ext;

    return filename.rfind(ext_compl) == (filename.size() - ext_compl.size());
    }

//-----------------------------------------------------------------------------
// Cut file extension

string CutExtension(const string &filename){

    const int pos = filename.rfind(".");

    return (pos > 0) ? filename.substr(0, pos) : filename;
    }

//-----------------------------------------------------------------------------
// Write file to stdout

bool PrintFile(const string &src_path){

    if(!src_path.empty()){
        ifstream src(src_path);
        if(src.good()){
            cout << src.rdbuf();
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Write data to a specified file starting from symbol with index 'start' end
// finishing with a symbol with index 'end'

bool WriteToFile(const string &fname, const string &data, int start, int end){

    // Create output stream for the file
    ofstream file(fname);

    // Default limits
    if(start < 0){
        start = 0;
        }
    if(end < 0 || end > (int)data.size()){
        end = data.size();
        }

    // If file is ready for writing
    if(file.good()){
        // If there is a data to write
        if(!data.empty()){
            for(int i = start; i < end; i++){
                file.put(data[i]);
                }
            }
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Copy 'src_path' file to 'dst_path' file

bool CopyFile(const string& src_path, const string& dst_path){

    if(!src_path.empty() && !dst_path.empty()){
        ifstream src(src_path);
        ofstream dst(dst_path);
        if(src.good() && dst.good()){
            dst << src.rdbuf();
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Format given file path

string FormatFilePath(string path, const int is_dir, const bool check_special_words){

    // Replacing special words at the beginning of a path
    if(!path.empty() && check_special_words){

        bool sw_initialized = false;
        map<string, string> special_words;
        special_words["~"];

        // Initialize the 'special_words' map
        auto initialize_sw = [&] (void){
            special_words["~"] = GetHomeDir();
            sw_initialized = true;
            };
        // - nested func

        int pos = 0;
        for(const auto &sw : special_words){
            pos = path.find(sw.first, pos);
            if(pos == 0){
                if(!sw_initialized){
                    initialize_sw();
                    }
                path.erase(pos, sw.first.size());
                path.insert(pos, sw.second);
                break;
                }
            }
        }

    // Path of a dir should ends with a slash symbol
    if(is_dir == IS_DIR){
        path = DirEnding(path);
        }
    else if(is_dir == NOT_A_DIR && !path.empty() && path.back()=='/'){
        path.pop_back();
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Get full path of a home directory

string GetHomeDir(void){

    const string homedir = getenv("HOME");

    return homedir.empty() ? getpwuid(getuid())->pw_dir : homedir;
    }

//-----------------------------------------------------------------------------
// Generate unique file name

string GenFileName(const string &base, const string &postfix){

    string filepath;

    for(int i = 1;; i++){
        filepath = base + postfix + to_string(i);
        // Checking file existence
        if(!CheckFile(filepath)){
            break;
            }
        }

    return filepath;
    }

//-----------------------------------------------------------------------------
// Complement path with a slash symbol

string DirEnding(const string &path){

    if(path.empty()){
        return "./";
        }

    if(path.back() != '/'){
        return path + "/";
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Strip path from filename

string StripPath(const string &path){

    const size_t n = path.rfind("/");

    if(n != string::npos){
        return (n == (path.size() - 1)) ? "" : path.substr(n + 1);
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Check that given dir name is a reference dir name {".", "./", "..", "../"}

bool DotsDir(const string& dir_name){

    if(!dir_name.empty() && dir_name.at(0) == '.'){

        if(dir_name.size() == 1){
            return true;
            }

        if(dir_name.at(1) == '/'){
            return(dir_name.size() == 2);
            }

        if(dir_name.at(1) == '.'){
            if(dir_name.size() == 2){
                return true;
                }
            if(dir_name.at(2) == '/'){
                return(dir_name.size() == 3);
                }
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Compare file names (dir name may be with or without '/' symbol at the end)

bool CompareFilenames(const string &name1, const string &name2){

    int i;
    bool name1_is_longer;

    if(name1.size() >= name2.size()){
        name1_is_longer = true;
        i = name1.find(name2);
        }
    else{
        name1_is_longer = false;
        i = name2.find(name1);
        }

    return (i == 0 && (
        name1.size() == name2.size() ||
        (name1_is_longer && name1.size() == (name2.size() + 1) && name1.back() == '/') ||
        (!name1_is_longer && name2.size() == (name1.size() + 1) && name2.back() == '/')
        ));
    }

//-----------------------------------------------------------------------------
// Join file paths

string JoinFilePaths(const string &path_left, const string &path_right){

    if(!path_left.empty() && !path_right.empty()){
        if(path_left.back() == '/'){
            if(path_right.front() == '/'){
                return path_left + path_right.substr(1);
                }
            }
        else{
            if(path_right.front() != '/'){
                return path_left + "/" + path_right;
                }
            }
        }

    return path_left + path_right;
    }

//-----------------------------------------------------------------------------
// Relocate given file paths to a specified base directory

vector<string> RebaseFilePaths(const vector<string> &paths,
        const string &base_dir){

    if(base_dir.empty()){
        return paths;
        }

    vector<string> rebased_paths;

    for(auto &p: paths){
        rebased_paths.push_back(JoinFilePaths(base_dir, p));
        }

    return rebased_paths;
    }

//-----------------------------------------------------------------------------
// Check that given file path is present in a given list of file paths or that
// it is a subpath of at least one of the paths of that list

bool FileIsSomewhereInThere(const string &fname, const vector<string> &paths){

    if(paths.empty() || fname.empty()){
        return false;
        }

    const string fname_formatted = FormatFilePath(fname, IS_DIR);

    for(const auto &p : paths){
        if(p.empty()){
            continue;
            }
        const string path_formatted = FormatFilePath(p, IS_DIR);
        if(fname_formatted.size() >= path_formatted.size() &&
                fname_formatted.compare(0, path_formatted.size(), path_formatted) == 0){
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// A version of the "FileIsSomewhereInThere" function with a second param being
// a single path instead of a vector of paths

bool FileIsSomewhereInThere(const string &fname, const string &path){

    return FileIsSomewhereInThere(fname, vector<string>(1, path));
    }

//-----------------------------------------------------------------------------
// Fast version of the "FileIsSomewhereInThere" function without formatting of
// the paths

bool FileIsInThere(const string &fname, const vector<string> &paths){

    if(paths.empty() || fname.empty()){
        return false;
        }

    const bool fname_ends_with_slash = fname.back() == '/';

    for(const auto &path : paths){
        if(path.empty()){
            continue;
            }
        if(fname_ends_with_slash){
            if(path.back() == '/'){
                if(fname.size() >= path.size() &&
                        fname.compare(0, path.size(), path) == 0){
                    return true;
                    }
                }
            else{
                if(fname.size() >= (path.size() + 1) &&
                        fname.at(path.size()) == '/' &&
                        fname.compare(0, path.size(), path) == 0){
                    return true;
                    }
                }
            }
        else{
            if(path.back() == '/'){
                if(fname.size() == (path.size() - 1)){
                    if(fname.compare(0, path.size() - 1, path) == 0){
                        return true;
                        }
                    }
                else{
                    if(fname.size() > (path.size() - 1) &&
                            fname.compare(0, path.size(), path) == 0){
                        return true;
                        }
                    }
                }
            else{
                if(fname.size() == path.size()){
                    if(fname == path){
                        return true;
                        }
                    }
                else{
                    if(fname.size() > path.size() &&
                            fname.at(path.size()) == '/' &&
                            fname.compare(0, path.size(), path) == 0){
                        return true;
                        }
                    }
                }
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Go up in a file path

string Backspace(string path){

    path = DirEnding(path);
    path.pop_back();

    const auto i = path.rfind("/");

    if(i == string::npos){
        return "";
        }

    return path.substr(0, i + 1);
    }

//-----------------------------------------------------------------------------
// Open specified directory and call given function, which should write the
// result to the "result" string; The "err" string is used to pass back an
// error message

bool OpenAndPerform(const string &dir_path,
        function<bool(const string &dp, DIR *dir, string &res)> func,
        string &result, string &err){

    bool flag = true;
    DIR *dir;

    if((dir = opendir(dir_path.c_str()))){
        if(!func(dir_path, dir, result)){
            err = "Can't apply given function to the "
                "\"" + dir_path + "\" directory";
            flag = false;
            }
        closedir(dir);
        }
    else{
        err = "Can't open \"" + dir_path + "\" directory";
        flag = false;
        }

    return flag;
    }

//-----------------------------------------------------------------------------
// Open specified directory and call given function for all files of that dir

bool OpenAndPerformOnAllFiles(const string &dir_path,
        function<bool(const string &dp, dirent *ent)> func,
        string &err){

    bool flag = true;
    DIR *dir;

    if((dir = opendir(dir_path.c_str()))){
        // For all files of the dir
        dirent *f_ent;
        const string dir_path_fin = DirEnding(dir_path);
        while((f_ent = readdir(dir))){
            const string f_path = dir_path_fin + f_ent->d_name;
            if(!func(f_path, f_ent)){
                err = "Failed operation for the \"" + f_path + "\" file";
                flag = false;
                break;
                }
            }
        closedir(dir);
        }
    else{
        err = "Could not open " + dir_path + "directory";
        flag = false;
        }

    return flag;
    }

//-----------------------------------------------------------------------------
// Check that file is a directory

bool IsDir(const string &path){

    struct stat path_stat;

    if(stat(path.c_str(), &path_stat) < 0){
        return false;
        }

    return S_ISDIR(path_stat.st_mode);
    }

//-----------------------------------------------------------------------------
// Read file to memory

bool TFileInMem::ReadFileToMem(const string &file_path){

    // Check existence of a file
    if(!CheckFile(file_path)){
        return false;
        }

    // Save file path
    fname.assign(file_path);
    // Create input file stream
    ifstream file(file_path);
    // Write to memory from the file stream
    data.assign((istreambuf_iterator<char>(file)),
        (istreambuf_iterator<char>()));

    return true;
    }

//-----------------------------------------------------------------------------
// Get index of a 1-st symbol of a n-th line placed after the 'start' index

int TFileInMem::NLineIndex(const int n, const int start) const{

    int pos, pos0;
    pos = pos0 = start;

    for(int k = 0; k < n; k++){
        // Searching for an index of a next line or the '\0' symbol
        pos = data.find('\n', pos);
        if(pos < 0){
            pos = data.find('\0', pos0);
            }
        // Skipping the '\0' symbol
        pos++;
        pos0 = pos;
        }

    return pos;
    }

//-----------------------------------------------------------------------------
// Sort lines of the file in alphabetical order

void TFileInMem::SortLines(void){

    vector<string> lines;
    string line;

    // Deconstructing the text to a vector of lines
    for(auto c = data.begin(); c != data.end(); c++){
        const bool newline = (*c == '\n');
        if(!newline){
            line.push_back(*c);
            }
        if(newline || c == (data.end() - 1)){
            lines.push_back(line);
            line.clear();
            }
        }

    // Sorting the vector of lines
    sort(lines.begin(), lines.end(),
        [](const string& l, const string& r) -> bool{return r > l;});

    // Constructing the text back again, this time with sorted lines
    data.clear();
    for(auto &l: lines){
        data += l + "\n";
        }
    }

//-----------------------------------------------------------------------------
