/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef lcontrolpp_filesystemH
#define lcontrolpp_filesystemH

#include "libcontrol++.h"
#include <vector>
#include <functional>
#include <dirent.h>

#define NOT_A_DIR   0       // File is not a dir
#define IS_DIR      1       // File is a dir

//-----------------------------------------------------------------------------

// Check file existence
bool CheckFile(const string &path);
// Check file extension
bool CheckExtension(const string &filename, const string &ext);
// Cut file extension
string CutExtension(const string &filename);
// Write file to stdout
bool PrintFile(const string &src_path);
// Write data to a specified file starting from symbol with index 'start' end
// finishing with a symbol with index 'end'
bool WriteToFile(const string &fname, const string &data, int start=-1,
        int end=-1);
// Copy 'src_path' file to 'dst_path' file
bool CopyFile(const string &src_path, const string &dst_path);
// Format given file path
string FormatFilePath(string path, const int is_dir=-1,
        const bool check_special_words=true);
// Get full path of a home directory
string GetHomeDir(void);
// Generate unique file name
string GenFileName(const string &base, const string &postfix="");
// Strip path from filename
string StripPath(const string &path);
// Complement path with a slash symbol
string DirEnding(const string &path);
// Check that given dirname is a reference dir name {".", "./", "..", "../"}
bool DotsDir(const string &dir_name);
// Compare file names (dir name may be with or without '/' symbol at the end)
bool CompareFilenames(const string &name1, const string &name2);
// Relocate given file paths to a specified base directory
vector<string> RebaseFilePaths(const vector<string> &paths,
        const string &base_dir);
// Check that given file path is present in a given list of file paths or that
// it is a subpath of at least one of the paths of that list
bool FileIsSomewhereInThere(const string &fname, const vector<string> &paths);
bool FileIsSomewhereInThere(const string &fname, const string &path);
bool FileIsInThere(const string &fname, const vector<string> &paths);
// Go up in a file path
string Backspace(string path);
// Open specified directory and call given function
bool OpenAndPerform(const string &dir_path,
        function<bool(const string &dp, DIR *dir, string &res)> func,
        string &result, string &err);
// Open specified directory and call given function for all files of that dir
bool OpenAndPerformOnAllFiles(const string &dir_path,
        function<bool(const string &dp, dirent *ent)> func,
        string &err);
// Check that file is a directory
bool IsDir(const string &path);

//-----------------------------------------------------------------------------
// Join file paths

string JoinFilePaths(const string &path_left, const string &path_right);

template<typename T> string JoinFilePaths(const T &paths){

    if(paths.empty()){
        return "";
        }

    string path = paths.front();
    for(auto p = (paths.begin() + 1); p != paths.end(); p++){
        path = JoinFilePaths(path, *p);
        }

    return path;
    }

//-----------------------------------------------------------------------------
// File in memory

class TFileInMem{
    private:
    public:
        string fname;  // File full path
        string data;   // File data
        // Access data
        const char& operator[](const int i) const{return data.at(i);}
        // Read file to the memory
        bool ReadFileToMem(const string &file_path);
        // Write part of the data back to the fname
        bool WriteBackToFile(const int start=-1, const int end=-1) const{
            return WriteToFile(fname, data, start, end);
            }
        // Get index of the 1-st symbol of the n-th line placed after
        // the 'start' index
        int NLineIndex(const int n, const int start=0) const;
        // Sort lines of the file in alphabetical order
        void SortLines(void);

        TFileInMem(void){}
        ~TFileInMem(void){}
    };

//-----------------------------------------------------------------------------

#endif
