/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "permissions_basic.h"
#include <algorithm>
#include <sys/stat.h>

const array<char, 3> rwx = {'r', 'w', 'x'};
const array<char, 7> rwx_full = {'r', 'w', 'x', 's', 'S', 't', 'T'};
const array<char, 3> ugo = {'u', 'g', 'o'};

//-----------------------------------------------------------------------------
// Write ID of a specified user

bool GetUserId(uid_t &uid, const string &user, string &err){

    // Pointer to a struct that can be used to get information about a user
    // from /etc/passwd file
    const struct passwd *pwd = getpwnam(user.c_str());
    if(!pwd){
        err = "Can't find stated user \"" + user + "\"";
        return false;
        }

    uid = pwd->pw_uid;

    return true;
    }

//-----------------------------------------------------------------------------
// Write ID of a specified group

bool GetGroupId(gid_t &gid, const string &group, string &err){

    // Pointer to a struct that can be used to get information about a group
    // from /etc/group file
    const struct group *grp = getgrnam(group.c_str());
    if(!grp){
        err = "Can't find stated group \"" + group + "\"";
        return false;
        }

    gid = grp->gr_gid;

    return true;
    }

//-----------------------------------------------------------------------------
// Get name of a specified user

string GetUserName(const uid_t uid, string &err){

    const struct passwd *pwd = getpwuid(uid);

    if(!pwd){
        err = "Can't find user with id " + uid;
        return "";
        }

    return pwd->pw_name;
    }

//-----------------------------------------------------------------------------
// Get name of a specified group

string GetGroupName(const gid_t gid, string &err){

    const struct group *grp = getgrgid(gid);

    if(!grp){
        err = "Can't find group with id " + gid;
        return "";
        }

    return grp->gr_name;
    }

//-----------------------------------------------------------------------------
// Verify owner of a file that is represented by a stat struct

bool VerifyOwner(const struct stat &st, const string &owner){

    if(owner != "*"){
        const struct passwd *current_pwd = getpwuid(st.st_uid);
        const struct passwd *pwd = getpwnam(owner.c_str());
        if(!current_pwd || !pwd || pwd->pw_uid != current_pwd->pw_uid){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Verify group of a file that is represented by a stat struct

bool VerifyGroup(const struct stat &st, const string &group){

    if(group != "*"){
        const struct group *current_grp = getgrgid(st.st_gid);
        const struct group *grp = getgrnam(group.c_str());
        if(!current_grp || !grp || grp->gr_gid != current_grp->gr_gid){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Set according bit of a mode, where 'new_state' is a bit to be set

void SetPermModeBit(struct stat &st, const char new_state, const char rwx,
        const char ugo){

    // '*' symbol means "don't change that bit"
    if(new_state == '*'){
        return;
        }

    if(IsRWX(new_state)){
        st.st_mode |= GetPermModeMask(new_state, ugo);
        }
    else{
        st.st_mode &= (~GetPermModeMask(rwx, ugo, true));
        }
    }

//-----------------------------------------------------------------------------
// Get bit mask with single 1 bit for a given rwx & ugo pair

mode_t GetPermModeMask(const char rwx, const char ugo,
        const bool with_spec_bit){

    switch(rwx){

        case 'r':
            if(ugo == 'u')
                return S_IRUSR;
            else if(ugo == 'g')
                return S_IRGRP;
            return S_IROTH;

        case 'w':
            if(ugo == 'u')
                return S_IWUSR;
            else if(ugo == 'g')
                return S_IWGRP;
            return S_IWOTH;

        default:

            switch(ugo){

                case 'u':
                    if(rwx == 's' || with_spec_bit)
                        return S_ISUID | S_IXUSR;
                    if(rwx == 'S')
                        return S_ISUID;
                    return S_IXUSR;

                case 'g':
                    if(rwx == 's' || with_spec_bit)
                        return S_ISGID | S_IXGRP;
                    if(rwx == 'S')
                        return S_ISGID;
                    return S_IXGRP;

                case 'o':
                    if(rwx == 't' || with_spec_bit)
                        return S_ISVTX | S_IXOTH;
                    if(rwx == 'T')
                        return S_ISVTX;
                    return S_IXOTH;
                }
        }

    return 0;
    }

//-----------------------------------------------------------------------------
// Compare value of a specified bit of a given permission mode with a given
// value of a mode bit

bool ComparePermModeBit(const mode_t mode, const char b, const char rwx,
        const char ugo){

    const mode_t mask = GetPermModeMask(rwx, ugo);

    return ((mode & mask) && IsRWX(b)) || (!(mode & mask) && !IsRWX(b));
    }

//-----------------------------------------------------------------------------
// Compare permission modes, where one mode is represented by mode_t var and
// another one is represented by 'rwxrwxrwx' type of string

bool ComparePermModes(const mode_t mode_a, const string &mode_b){

    if(!CheckModeString(mode_b)){
        return false;
        }

    int index = -1;

    for(const auto identity : ugo){
        for(const auto action : rwx){
            index++;
            if(mode_b.at(index) == '*'){
                continue;
                }
            if(!ComparePermModeBit(mode_a, mode_b.at(index), action, identity)){
                return false;
                }
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Compare permission modes represented by 'rwxrwxrwx' type of strings

bool ComparePermModes(const string &mode_a, const string &mode_b){

    if(!CheckModeString(mode_a) || !CheckModeString(mode_b)){
        return false;
        }

    for(int i = 0; i < 9; i++){
        if(mode_a.at(i) == '*' || mode_b.at(i) == '*' ||
                mode_a.at(i) == mode_b.at(i)){
            continue;
            }
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Get 'ls -l' type of mode string from a given t_mode

string GetPermModeString(const mode_t mode){

    string mode_str;

    mode_str  = (mode & S_IRUSR) ? "r" : "-";
    mode_str += (mode & S_IWUSR) ? "w" : "-";
    mode_str += XPerm(mode, S_IXUSR, S_ISUID);
    mode_str += (mode & S_IRGRP) ? "r" : "-";
    mode_str += (mode & S_IWGRP) ? "w" : "-";
    mode_str += XPerm(mode, S_IXGRP, S_ISGID);
    mode_str += (mode & S_IROTH) ? "r" : "-";
    mode_str += (mode & S_IWOTH) ? "w" : "-";
    mode_str += XPerm(mode, S_IXOTH, S_ISVTX);

    return mode_str;
    }

//-----------------------------------------------------------------------------
// Get a type of exe permission (with/without suid/sticky bit)

char XPerm(const mode_t mode, const mode_t exe_bit, const mode_t spec_bit){

    const bool exe = mode & exe_bit;

    switch(spec_bit){

        case S_ISUID:
            if(mode & S_ISUID) return exe ? 's' : 'S';
            break;

        case S_ISGID:
            if(mode & S_ISGID) return exe ? 's' : 'S';
            break;

        case S_ISVTX:
            if(mode & S_ISVTX) return exe ? 't' : 'T';
            break;
        }

    return exe ? 'x' : '-';
    }

//-----------------------------------------------------------------------------
// Check that permission bit is read, wright or exe bit

bool IsRWX(const char b){

    return find(rwx_full.begin(), rwx_full.end(), b) != rwx_full.end();
    }

//-----------------------------------------------------------------------------
// Check format of a mode string

bool CheckModeString(const string &mode){

    if(mode.size() != 9){
        return false;
        }

    for(int i = 0; i < 9; i += 3){
        if(!CheckModeSection(mode.begin() + i, ugo.at(i/3))){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check format of a given rwx section of a mode string

bool CheckModeSection(const string::const_iterator mode_sec, const char ugo){

    for(auto b = mode_sec; b != (mode_sec + 3); b++){

        if(b == mode_sec){
            if(*b == 'r')
                continue;
            }
        else if(b == (mode_sec + 1)){
            if(*b == 'w')
                continue;
            }
        else if(b == (mode_sec + 2)){
            if(*b == 'x' ||
                    ((ugo == 'u' || ugo == 'g') && (*b == 's' || *b == 'S')) ||
                    (ugo == 'o' && (*b == 't' || *b == 'T')))
                continue;
            }

        if(*b == '-' || *b == '*')
            continue;

        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------

// Update ACL permission string
bool UpdateACLPermStr(string &perm, const acl_permset_t permset,
        const acl_perm_t type, string &err){

    if(find(acl_perm_types.begin(), acl_perm_types.end(), type) ==
            acl_perm_types.end()){
        err = "Wrong ACL permission type";
        return false;
        }

    if((int)perm.size() != 3){
        err = "ACL permission should consist of 3 symbols";
        return false;
        }

    // Return value 1 means 'allowed', 0 means 'disallowed' and -1 means 'error'
    const int perm_val = acl_get_perm(permset, type);
    if(perm_val == -1){
        err = "Can't get ACL permission bit";
        return false;
        }

    if(type == ACL_READ){
        perm[0] = (perm_val == 1) ? 'r' : '-';
        }
    else if(type == ACL_WRITE){
        perm[1] = (perm_val == 1) ? 'w' : '-';
        }
    else{
        perm[2] = (perm_val == 1) ? 'x' : '-';
        }

    return true;
    }

//-----------------------------------------------------------------------------
