/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef lcontrolpp_permissions_basicH
#define lcontrolpp_permissions_basicH

#include "libcontrol++.h"

#include <unistd.h>         // Contains "chown" function description
#include <pwd.h>            // Contains "passwd" struct description
#include <grp.h>            // Contains "group" struct description
#include <sys/acl.h>        // Contains "acl_*" functions
#include <acl/libacl.h>
#include <array>
#include <map>

#define ALLOW_EXE       "**x**x**x" // Mode string to allow execution
#define DISALLOW_EXE    "**-**-**-" // Mode string to disallow execution
#define DONT_CHANGE     "*********" // Unvarying mode string
#define DISALLOW_ALL    "---------" // Unvarying mode string

//-----------------------------------------------------------------------------

// rwx - 'r' (read), 'w' (write) or 'x' (execute) mode
extern const array<char, 3> rwx;
// uqo - 'u' (user), 'g' (group) or 'o' (others) attachment
extern const array<char, 3> ugo;
// Read, write or execute ACL permission types
extern const array<acl_perm_t, 3> acl_perm_types;

//-----------------------------------------------------------------------------

// Write ID of a specified user
bool GetUserId(uid_t &uid, const string &user, string &err);
// Write ID of a specified group
bool GetGroupId(gid_t &gid, const string &grp, string &err);

//-----------------------------------------------------------------------------

// Get name of a specified user
string GetUserName(const uid_t uid, string &err);
//Get name of a specified group
string GetGroupName(const gid_t gid, string &err);

//-----------------------------------------------------------------------------

// Verify owner of a file that is represented by a stat struct
bool VerifyOwner(const struct stat &st, const string &owner);
// Verify group of a file that is represented by a stat struct
bool VerifyGroup(const struct stat &st, const string &group);

//-----------------------------------------------------------------------------

// Set according bit of a mode, where 'new_state' is a bit to be set
void SetPermModeBit(struct stat &st, const char new_state, const char rwx,
    const char ugo);
// Get bit mask with single 1 bit for a given rwx & ugo pair
mode_t GetPermModeMask(const char rwx, const char ugo,
    const bool with_spec_bit=false);

//-----------------------------------------------------------------------------

// Compare value of a specified bit of a given permission mode with a given
// value of a mode bit
bool ComparePermModeBit(const mode_t st, const char rwx, const char ugo,
    const char b);
// Compare permission modes, where one mode is represented by mode_t var and
// another one is represented by 'rwxrwxrwx' type of string
bool ComparePermModes(const mode_t mode_a, const string &mode_b);
// Compare permission modes represented by 'rwxrwxrwx' type of strings
bool ComparePermModes(const string &mode_a, const string &mode_b);

//-----------------------------------------------------------------------------

// Get 'ls -l' type of mode string from a given t_mode
string GetPermModeString(const mode_t mode);
// Get a type of exe permission (with/without suid/sticky bit)
char XPerm(const mode_t mode, const mode_t exe_bit, const mode_t spec_bit);
// Check that permission bit is read, wright or exe bit
bool IsRWX(const char b);
// Check format of a mode string
bool CheckModeString(const string &mode);
// Check format of a given rwx section of a mode string
bool CheckModeSection(const string::const_iterator mode_sec,
    const char ugo='\0');

//-----------------------------------------------------------------------------

// Update ACL permission string
bool UpdateACLPermStr(string &perm, const acl_permset_t permset,
    const acl_perm_t type, string &err);

//-----------------------------------------------------------------------------

#endif
