/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "permissions_file_acl.h"
#include "text.h"
#include <sys/stat.h>

const array<acl_perm_t, 3> acl_perm_types = {ACL_READ, ACL_WRITE, ACL_EXECUTE};

//-----------------------------------------------------------------------------
// Get current ACL of a file

const string acl_free_err = "Can't free acl memory";

TFileACL GetACL(const string &file_path, string &err){

    TFileACL res_acl;

    // Return empty ACL obj if file not exist
    if(access(file_path.c_str(), F_OK) == -1){
        return res_acl;
        }

    // Get current ACL of a file (ACL_TYPE_ACCESS is current ACL,
    // ACL_TYPE_DEFAULT is default ACl)
    const acl_t acl = acl_get_file(file_path.c_str(), ACL_TYPE_ACCESS);
    if(!acl){
        err = "Can't get ACL of a " + file_path + " file";
        return TFileACL();
        }
    res_acl.WritePath(file_path);

    acl_entry_t entry;

    // For each entry of this ACL
    for(int entry_id = ACL_FIRST_ENTRY;
            acl_get_entry(acl, entry_id, &entry) == 1;
            entry_id = ACL_NEXT_ENTRY){

        acl_tag_t tag;

        // Retrieve tag type
        if(acl_get_tag_type(entry, &tag) == -1){
            err = "Can't get ACL tag";
            return TFileACL();
            }

        string qualifier;

        // Retrieve optional user tag qualifier
        if(tag == ACL_USER){
            uid_t *user_id = (uid_t*)acl_get_qualifier(entry);
            if(!user_id){
                err = "Can't get ACL user qualifier";
                return TFileACL();
                }
            qualifier = GetUserName(*user_id, err);
            if(acl_free(user_id) == -1){
                err = acl_free_err;
                return TFileACL();
                }
            }
        else if(tag == ACL_GROUP){
            gid_t *group_id = (gid_t*)acl_get_qualifier(entry);
            if(!group_id){
                err = "Can't get ACL group qualifier";
                return TFileACL();
                }
            qualifier = GetGroupName(*group_id, err);
            if(acl_free(group_id) == -1){
                err = acl_free_err;
                return TFileACL();
                }
            }

        string perm(3, '-');
        acl_permset_t permset;

        // Retrieve and write permissions
        if(acl_get_permset(entry, &permset) == -1){
            err = "Can't get ACL permset";
            return TFileACL();
            }
        if(UpdateACLPermStr(perm, permset, ACL_READ, err) &&
                UpdateACLPermStr(perm, permset, ACL_WRITE, err) &&
                UpdateACLPermStr(perm, permset, ACL_EXECUTE, err)){
            // Write current ACL param
            res_acl.WriteByTag(tag, perm, qualifier);
            }
        else{
            return TFileACL();
            }
        }

    if(acl_free(acl) == -1){
        err = acl_free_err;
        return TFileACL();
        }

    return res_acl;
    }

//-----------------------------------------------------------------------------
// ACL of a file

TFileACL::TFileACL(const string &p, const string &u, const string &g,
        const string &m, const string &o){

    WriteNecessary(p, u, g, m, o);
    p_attrs = &attrs;
    }


TFileACL::TFileACL(const string &p, const string &u,
        const map<string, string> &u_q, const string &g,
        const map<string, string> &g_q, const string &m,
        const string &o){

    WriteEverything(p, u, u_q, g, g_q, m, o);
    p_attrs = &attrs;
    }


TFileACL::TFileACL(const map<string, string> *attributes){

    p_attrs = attributes;
    }

//-----------------------------------------------------------------------------
// Copy ACL object

TFileACL::TFileACL(const TFileACL &f){

    CopyFileACL(f);
    }


TFileACL& TFileACL::operator=(const TFileACL &f){

    CopyFileACL(f);
    return *this;
    }


void TFileACL::CopyFileACL(const TFileACL &f){

    WriteEverything(f.Path(), f.User(), f.QUsers(), f.Group(), f.QGroups(),
        f.Mask(), f.Other());
    p_attrs = &attrs;
    }

//-----------------------------------------------------------------------------
// Write values of necessary ACl params

bool TFileACL::WriteNecessary(const string &p, const string &u, const string &g,
        const string &m, const string &o){

    return WritePath(p) && WriteUser(u) && WriteGroup(g) && WriteMask(m) &&
        WriteOther(o);
    }

//-----------------------------------------------------------------------------
// Write values of all ACL params

bool TFileACL::WriteEverything(const string &p, const string &u,
        const map<string, string> &u_q, const string &g,
        const map<string, string> &g_q, const string &m, const string &o){

    return WriteNecessary(p, u, g, m, o) && WriteQUsers(u_q) &&
        WriteQGroups(g_q);
    }

//-----------------------------------------------------------------------------
// Modify specified data field

bool TFileACL::WriteAttr(string &str_attr, map<string, string> &map_attr,
        const string &attr_val, const string &q, const bool mode_str){

    if(attr_val.empty()){
        return false;
        }

    if(mode_str && (attr_val.size() < 3 || !CheckModeSection(attr_val.begin()))){
        return false;
        }

    if(!q.empty()){
        map_attr[q] = attr_val;
        }
    else{
        str_attr = attr_val;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Get main data fields

vector<string> TFileACL::GetData(void){

    vector<string> data(FACL_PARAMS_N);

    data[FACL_PATH]  = Path();
    data[FACL_USER]  = p_attrs->at("user");
    data[FACL_GROUP] = p_attrs->at("group");
    data[FACL_MASK]  = p_attrs->at("mask");
    data[FACL_OTHER] = p_attrs->at("other");

    return data;
    }

//-----------------------------------------------------------------------------
// Modifiers

bool TFileACL::WriteQUsers(const map<string, string> &u_q){

    for(const auto &u : u_q){
        if(WriteUser(u.second, u.first)){
            return false;
            }
        }

    return true;
    }


bool TFileACL::WriteQGroups(const map<string, string> &g_q){

    for(const auto &g : g_q){
        if(WriteGroup(g.second, g.first)){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Assign specified data field with given value

bool TFileACL::Assgn(const string &var, const string &val){

    if(var == "path"){
        return WritePath(val);
        }
    else if(var == "mask"){
        return WriteMask(val);
        }
    else if(var == "other"){
        return WriteOther(val);
        }
    else{
        string qualifier;
        const int q_index = var.find(":");
        if(q_index > 0 && (q_index + 1) < (int)var.size()){
            qualifier = var.substr(q_index + 1);
            }
        if(var.compare(0, 4, "user") == 0){
            return WriteUser(val, qualifier);
            }
        else if(var.compare(0, 5, "group") == 0){
            return WriteGroup(val, qualifier);
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Write ACL param identified by given tag

bool TFileACL::WriteByTag(const acl_tag_t tag, const string &val,
        const string &q){

    if(tag == ACL_USER_OBJ){
        return WriteUser(val);
        }
    else if(tag == ACL_USER){
        return WriteUser(val, q);
        }
    else if(tag == ACL_GROUP_OBJ){
        return WriteGroup(val);
        }
    else if(tag == ACL_GROUP){
        return WriteGroup(val, q);
        }
    else if(tag == ACL_MASK){
        return WriteMask(val);
        }
    else if(tag == ACL_OTHER){
        return WriteOther(val);
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Copy the mode without any checks

void TFileACL::CopyMode(const vector<string> &m){

    if(m.size() >= FACL_PARAMS_N){
        attrs["user"]  = m.at(FACL_USER);
        attrs["group"] = m.at(FACL_GROUP);
        attrs["mask"]  = m.at(FACL_MASK);
        attrs["other"] = m.at(FACL_OTHER);
        p_attrs = &attrs;
        }
    }


void TFileACL::CopyMode(const string &m){

    attrs["user"]  = m;
    attrs["group"] = m;
    attrs["mask"]  = m;
    attrs["other"] = m;
    p_attrs = &attrs;
    }

//-----------------------------------------------------------------------------
// Revise the mode

bool TFileACL::Revise(const TSectorType &sec, string &err){

    if(path.empty()){
        err = "File path is missing";
        return false;
        }

    WritePath(FormatFilePath(path));

    // Default params for whitelist
    if(sec.WhiteList() && sec.ACL()){
        if(p_attrs->at("user").empty()){
            WriteUser("rwx");
            }
        if(p_attrs->at("group").empty()){
            WriteGroup("r-x");
            }
        if(p_attrs->at("mask").empty()){
            WriteMask("r-x");
            }
        if(p_attrs->at("other").empty()){
            WriteOther("--x");
            }
        }
    // Default params for blacklist
    else if(sec.BlackList() && sec.ACL()){
        if(p_attrs->at("user").empty()){
            WriteUser("rw-");
            }
        if(p_attrs->at("group").empty()){
            WriteGroup("r--");
            }
        if(p_attrs->at("mask").empty()){
            WriteMask("r--");
            }
        if(p_attrs->at("other").empty()){
            WriteOther("---");
            }
        }
    // All basic ACL params must be defined for other types of sections
    else if(p_attrs->at("user").empty() || p_attrs->at("group").empty() ||
            p_attrs->at("mask").empty() || p_attrs->at("other").empty()){
        err = "Some of the basic ACL params are missing";
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Apply ACL and write new ACL to 'new_perm'

bool TFileACL::Apply(string &err, string *new_perm){
    return true;
    }

//-----------------------------------------------------------------------------
// Check compliance between assigned and currently set ACL of the file

bool TFileACL::CheckCompliance(string &err, string &warn){
    return true;
    }

//-----------------------------------------------------------------------------
// Check controversy with the given ACL

bool TFileACL::Controversy(const TFileACL &other_acl) const{

    return
        (p_attrs->at("user")  != "***" && other_acl.User()  != "***") ||
        (p_attrs->at("group") != "***" && other_acl.Group() != "***") ||
        (p_attrs->at("other") != "***" && other_acl.Other() != "***") ||
        (p_attrs->at("mask")  != "***" && other_acl.Mask()  != "***") ||
        (!user_q.empty() && !other_acl.QUsers().empty()) ||
        (!group_q.empty() && !other_acl.QGroups().empty());
    }

//-----------------------------------------------------------------------------
// Check that every data field is not empty

bool TFileACL::Complete(void) const{

    return !Path().empty() && !p_attrs->at("user").empty() &&
        !p_attrs->at("group").empty() && !p_attrs->at("other").empty() &&
        !p_attrs->at("mask").empty();
    }

//-----------------------------------------------------------------------------
// Check that mode can't have any effect

bool TFileACL::Neutral(void) const{

    return p_attrs->at("user") == "***" && p_attrs->at("group") == "***" &&
        p_attrs->at("other") == "***";
    }

//-----------------------------------------------------------------------------
// Print ACL params to a given output stream

void TFileACL::Print(ostream &out) const{

    out << "[file_acl]" << endl
        << "path = \""  << (CheckSymbols(path) ? Esc(path) : path) << "\"" << endl
        << "user = \""  << p_attrs->at("user") << "\"" << endl;

    for(const auto &u : user_q){
        out << "user:" << u.first  << " = \"" << u.second << "\"" << endl;
        }

    out << "group = \"" << p_attrs->at("group") << "\"" << endl;

    for(const auto &g : group_q){
        out << "group:" << g.first  << " = \"" << g.second << "\"" << endl;
        }

    out << "mask = \""  << p_attrs->at("mask")  << "\"" << endl
        << "other = \"" << p_attrs->at("other") << "\"" << endl;
    }

//-----------------------------------------------------------------------------
