/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef lcontrolpp_permissions_file_aclH
#define lcontrolpp_permissions_file_aclH

#include "permissions.h"

#define FACL_PARAMS_N   5
#define FACL_PATH       0
#define FACL_USER       1
#define FACL_GROUP      2
#define FACL_MASK       3
#define FACL_OTHER      4

//-----------------------------------------------------------------------------
// ACL of a file

class TFileACL : public TFileMode{

    protected:

        map<string, string> user_q;     // Qualified user
        map<string, string> group_q;    // Qualified group

        // Copy data of the permission object
        void CopyFileACL(const TFileACL &f);
        // Modify specified data field
        bool WriteAttr(string &str_attr, map<string, string> &map_attr,
                const string &attr_val, const string &q="",
                const bool mode_str=true);

    public:

        // Get the type id
        int Type(void) const{
            return TYPE_FACL;
            }

        // Fields access
        const string& User(void) const{
            return attrs.at("user");
            }
        const map<string, string>& QUsers(void) const{
            return user_q;
            }
        const string UserQ(const string &q) const{
            if(user_q.find(q) != user_q.end())
                return user_q.at(q);
            return "";
            }
        const string& Group(void) const{
            return attrs.at("group");
            }
        const map<string, string>& QGroups(void) const{
            return group_q;
            }
        const string GroupQ(const string &q) const{
            if(group_q.find(q) != group_q.end())
                return group_q.at(q);
            return "";
            }
        const string& Mask(void) const{
            return attrs.at("mask");
            }
        const string& Other(void) const{
            return attrs.at("other");
            }

        // Get main data fields
        vector<string> GetData(void);

        // Modifiers
        bool WriteUser(const string &val, const string &q=""){
            return WriteAttr(attrs["user"], user_q, val, q);
            }
        bool WriteGroup(const string &val, const string &q=""){
            return WriteAttr(attrs["group"], group_q, val, q);
            }
        bool WriteMask(const string &val){
            map<string, string> empty_map;
            return WriteAttr(attrs["mask"], empty_map, val);
            }
        bool WriteOther(const string &val){
            map<string, string> empty_map;
            return WriteAttr(attrs["other"], empty_map, val);
            }
        bool WriteQUsers(const map<string, string> &u_q);
        bool WriteQGroups(const map<string, string> &g_q);
        bool WriteNecessary(const string &p, const string &u, const string &g,
                const string &m, const string &o);
        bool WriteEverything(const string &p, const string &u,
                const map<string, string> &u_q, const string &g,
                const map<string, string> &g_q, const string &m,
                const string &o);

        // Write ACL param identified by given tag
        bool WriteByTag(const acl_tag_t tag, const string &val,
                const string &q="");

        void CopyMode(const vector<string> &m);
        void CopyMode(const string &m);
        bool Assgn(const string &var, const string &val);
        bool Revise(const TSectorType &sec, string &err);
        bool Apply(string &err, string *new_perm=NULL);
        bool CheckCompliance(string &err, string &warn);
        bool Controversy(const TFileACL &other_acl) const;
        bool Complete(void) const;
        bool Neutral(void) const;
        void Print(ostream &out) const;

        void Clear(void){
            user_q.clear();
            group_q.clear();
            TFileMode::Clear();
            }

        TFileACL(void){
            attrs["user"];
            attrs["group"];
            attrs["mask"];
            attrs["other"];
            p_attrs = &attrs;
            }
        TFileACL(const string &p, const string &u, const string &g,
                const string &m, const string &o);
        TFileACL(const string &p, const string &u,
                const map<string, string> &u_q, const string &g,
                const map<string, string> &g_q, const string &m,
                const string &o);
        TFileACL(const map<string, string> *attributes);
        TFileACL(const TFileACL &f);
        TFileACL &operator=(const TFileACL &f);

        ~TFileACL(void){}
        };

//-----------------------------------------------------------------------------

// Get current ACL of a file
TFileACL GetACL(const string &file_path, string &err);

//-----------------------------------------------------------------------------

#endif
