/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "permissions_file_perm.h"
#include "text.h"
#include <sys/stat.h>

//-----------------------------------------------------------------------------
// Get current permission of a file

TFilePerm GetPermission(const string &file_path, string &err){

    TFilePerm perm;
    struct stat file_stat;

    if(stat(file_path.c_str(), &file_stat) == (-1)){
        err = "Can't stat '" + file_path + "' file";
        }
    else{
        perm.WritePath(file_path) &&
        perm.WriteMode(GetPermModeString(file_stat.st_mode)) &&
        perm.WriteOwner(GetUserName(file_stat.st_uid, err)) &&
        perm.WriteGroup(GetGroupName(file_stat.st_gid, err));
        }

    return perm;
    }

//-----------------------------------------------------------------------------
// Permission for a particular file

TFilePerm::TFilePerm(const string &p, const string &o, const string &g,
        const string &m){

    WriteEverything(p, o, g, m);
    p_attrs = &attrs;
    }


TFilePerm::TFilePerm(const map<string, string> *attributes){

    p_attrs = attributes;
    }

//-----------------------------------------------------------------------------
// Copy a TFilePerm object

TFilePerm::TFilePerm(const TFilePerm &f){

    CopyFilePerm(f);
    }


TFilePerm& TFilePerm::operator=(const TFilePerm &f){

    CopyFilePerm(f);
    return *this;
    }


void TFilePerm::CopyFilePerm(const TFilePerm &f){

    WriteEverything(f.Path(), f.Owner(), f.Group(), f.Mode());
    p_attrs = &attrs;
    }

//-----------------------------------------------------------------------------
// Write values of all permission params

bool TFilePerm::WriteEverything(const string &p, const string &o,
        const string &g, const string &m){

    return WritePath(p) && WriteOwner(o) && WriteGroup(g) && WriteMode(m);
    }

//-----------------------------------------------------------------------------
// TFilePerm fields access

const string& TFilePerm::Owner(void) const{
    return p_attrs->at("owner");
    }

const string& TFilePerm::Group(void) const{
    return p_attrs->at("group");
    }

const string& TFilePerm::Mode(void) const{
    return p_attrs->at("mode");
    }

string TFilePerm::OwnerMode(void) const{
    return (Mode().size() == 9) ? Mode().substr(0, 3) : "";
    }

string TFilePerm::GroupMode(void) const{
    return (Mode().size() == 9) ? Mode().substr(3, 3) : "";
    }

string TFilePerm::OtherMode(void) const{
    return (Mode().size() == 9) ? Mode().substr(6, 3) : "";
    }

//-----------------------------------------------------------------------------
// Get main data fields

vector<string> TFilePerm::GetData(void){

    vector<string> data;

    data.push_back(Path());
    data.push_back(p_attrs->at("owner"));
    data.push_back(p_attrs->at("group"));
    data.push_back(p_attrs->at("mode"));

    return data;
    }

//-----------------------------------------------------------------------------
// Assign specified data field with given value

bool TFilePerm::Assgn(const string &var, const string &val){

    if(var == "path"){
        return WritePath(val);
        }
    else if(var == "owner"){
        return WriteOwner(val);
        }
    else if(var == "group"){
        return WriteGroup(val);
        }
    else if(var == "mode"){
        return WriteMode(val);
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Copy the mode without any checks

void TFilePerm::PullData(void){

    for(auto &attr : *p_attrs){
        attrs[attr.first] = attr.second;
        }
    p_attrs = &attrs;
    }


void TFilePerm::CopyMode(const vector<string> &m){

    if(m.size() >= FPERM_PARAMS_N){
        attrs["mode"] = m.at(FPERM_MODE);
        PullData();
        }
    }


void TFilePerm::CopyMode(const string &m){

    attrs["mode"] = m;
    PullData();
    }

//-----------------------------------------------------------------------------
// Revise the mode

bool TFilePerm::Revise(const TSectorType &sec, string &err){

    if(path.empty()){
        err = "File path is missing";
        return false;
        }

    WritePath(FormatFilePath(path));

    if(p_attrs->at("owner").empty()){
        WriteOwner("*");
        }

    if(p_attrs->at("group").empty()){
        WriteGroup("*");
        }

    if(p_attrs->at("mode").empty()){
        // Default mode for blacklist
        if(sec.BlackList()){
            WriteMode(DISALLOW_EXE);
            }
        // Default mode for every other section types
        else{
            WriteMode(DONT_CHANGE);
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Apply the mode and write it's string representation to 'new_perm'

bool TFilePerm::Apply(string &err, string *new_perm){

    TFilePerm* np = NULL;

    if(new_perm){
        np = new TFilePerm;
        }

    bool success = ChangeOwnerAndGroup(err, np) && ChangePermMode(err, np);

    if(np){
        *new_perm =
            np->Mode()  + " " +
            np->Owner() + " " +
            np->Group() + " " +
            path;
        delete np;
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Chown ang chgrp of the file and write params of a new mode to 'new_perm' obj

bool TFilePerm::ChangeOwnerAndGroup(string &err, TFilePerm *new_perm){

    // New owner and new group ("-1" means "don't change")
    uid_t new_owner = -1;
    gid_t new_group = -1;

    // "*" symbol means "don't change the group/owner"
    if(p_attrs->at("owner") == "*" && p_attrs->at("group") == "*"){
        goto ChangeOwnerAndGroupReturn;
        }

    // If owner must be changed
    if(p_attrs->at("owner") != "*"){
        if(!GetUserId(new_owner, p_attrs->at("owner"), err)){
            return false;
            }
        }

    // If group must be changed
    if(p_attrs->at("group") != "*"){
        if(!GetGroupId(new_group, p_attrs->at("group"), err)){
            return false;
            }
        }

    // Trying to change owner and group of the file (chown returns -1 on errors)
    if(chown(path.c_str(), new_owner, new_group) == (-1)){
        err = "Can't change ownership of stated file \"" + path + "\"";
        return false;
        }

    ChangeOwnerAndGroupReturn:

    // Reporting back a new user name and a new group name
    if(new_perm){
        struct stat file_stat;
        if(new_owner == (uid_t)(-1) || new_group == (gid_t)(-1)){
            if(stat(path.c_str(), &file_stat) == (-1)){
                err = "Can't stat file " + path;
                return false;
                }
            }
        new_perm -> WriteOwner(GetUserName(
            new_owner == (uid_t)(-1) ? file_stat.st_uid : new_owner, err)
            );
        new_perm -> WriteGroup(GetGroupName(
            new_group == (gid_t)(-1) ? file_stat.st_gid : new_group, err)
            );
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Chmod of the file

bool TFilePerm::ChangePermMode(string &err, TFilePerm *new_perm){

    // Get struct that contents information about the file
    // to determine current file mode
    struct stat file_stat;
    if(stat(path.c_str(), &file_stat) == (-1)){
        err = "Can't stat file " + path;
        return false;
        }

    const mode_t current_mode = file_stat.st_mode;

    // Setting according bits of the mode field of the file stat structure
    int index = 0;
    for(const auto &identity : ugo){
        for(const auto &action : rwx){
            SetPermModeBit(file_stat, p_attrs->at("mode")[index++],
                action, identity);
            }
        }

    // If new mode is the same as the current one
    if(file_stat.st_mode == current_mode){
        goto ChangePermModeReturn;
        }

    // Applying new mode
    if(chmod(path.c_str(), file_stat.st_mode) == (-1)){
        err = "Can't change mode for \"" + path + "\"";
        return false;
        }

    ChangePermModeReturn:

    // Reporting back a new mode in rwx format
    if(new_perm){
        new_perm -> WriteMode(GetPermModeString(file_stat.st_mode));
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check compliance between assigned and currently set mode of the file

bool TFilePerm::CheckCompliance(string &err, string &warn){

    auto CatErr = [&](const string &msg){
        err += (err.empty() ? "Different " : ", ") + msg;
        };

    err.clear();
    warn.clear();

    // Skipping if file is not exist
    if(access(path.c_str(), F_OK) == -1){
        warn = "File not exist";
        return true;
        }

    // Getting stat of the file
    struct stat current_st;
    if(stat(path.c_str(), &current_st) == (-1)){
        err = "Can't stat file";
        return false;
        }

    // Checking the owner
    if(!VerifyOwner(current_st, p_attrs->at("owner"))){
        CatErr("owner");
        }

    // Checking the group
    if(!VerifyGroup(current_st, p_attrs->at("group"))){
        CatErr("group");
        }

    // Checking the mode
    if(!ComparePermModes(current_st.st_mode, p_attrs->at("mode"))){
        CatErr("mode");
        }

    return err.empty();
    }

//-----------------------------------------------------------------------------
// Check controversy with a given mode

bool TFilePerm::Controversy(const TFilePerm &other_perm) const{

    return
        (p_attrs->at("mode")  != DONT_CHANGE && other_perm.Mode() != DONT_CHANGE) ||
        (p_attrs->at("owner") != "*" && other_perm.Owner() != "*") ||
        (p_attrs->at("group") != "*" && other_perm.Group() != "*");
    }

//-----------------------------------------------------------------------------
// Check that every data field is not empty

bool TFilePerm::Complete(void) const{

    return !Path().empty() && !Owner().empty() && !Group().empty() &&
        !Mode().empty();
    }

//-----------------------------------------------------------------------------
// Check that mode can't have any effect

bool TFilePerm::Neutral(void) const{

    return Owner() == "*" && Group() == "*" && Mode() == DONT_CHANGE;
    }

//-----------------------------------------------------------------------------
// Print permission mode in the config style

void TFilePerm::Print(ostream &out) const{

    out << "[file]" << endl
        << "path = \""  << (CheckSymbols(Path()) ? Esc(Path()) : Path()) << "\"" << endl
        << "owner = \"" << Owner() << "\"" << endl
        << "group = \"" << Group() << "\"" << endl
        << "mode = \""  << Mode()  << "\"" << endl;
    }

//-----------------------------------------------------------------------------
