/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "print.h"
#include <sys/ioctl.h>  // Terminal ...
#include <unistd.h>     // ... size

//-----------------------------------------------------------------------------
// Get terminal width

int GetTerminalWidth(void){

    struct winsize ws;

    ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws);

    return (ws.ws_col > TERM_MAX_WIDTH) ? TERM_MAX_WIDTH : ws.ws_col;
    }

//-----------------------------------------------------------------------------
// Print according to current mode

void TPrinter::Print(const string &str) const{

    if(silent){
        return;
        }

    cout << str;
    }

//-----------------------------------------------------------------------------
// Print line of extra info

void TPrinter::PrintExtraLine(const string &ex) const{

    Print("\t- " + ex + "\n");
    }

//-----------------------------------------------------------------------------
// Print joined indentation units

void TPrinter::PrintIndent(void) const{

    for(const auto &i : indent){
        Print(i);
        }
    }

//-----------------------------------------------------------------------------
// Get a full size of concatenated indentation strings

size_t TPrinter::IndentSize(void) const{

    size_t size = 0;

    for(const auto &i : indent){
        size += i.size();
        }

    return size;
    }

//-----------------------------------------------------------------------------
// Print left part and right part of the text tied with the 'tie' on entire
// terminal line

void TPrinter::CompleteTheLine(const string &status){

    if(silent){
        return;
        }

    if(!operation.empty()){
        if(plain){
            Print(" ");
            }
        else{
            int term_width = GetTerminalWidth();
            int tie_len = term_width-operation.size() - status.size() -
                IndentSize();
            while(tie_len < 0){
                tie_len += term_width;
                }
            string tie;
            for(int i = 0; i < tie_len; i++){
                tie += tie_symbol;
                }
            if(!tie.empty()){
                Print(tie);
                };
            }
        }
    else{
        // Indent was not printed because 'operation' is empty string
        PrintIndent();
        }

    Print(status + "\n");

    // Displaying extra info
    for(const auto &ex : extra){
        if(ex.empty()){
            continue;
            }
        PrintExtraLine(ex);
        }
    }

//-----------------------------------------------------------------------------
// Form special message

string TPrinter::SpecMessage(const string &preface, const string &msg) const{

    return ("[" + preface + (msg.empty() ? "" : (": " + msg)) + "]");
    }

//-----------------------------------------------------------------------------
// Start a new status line

void TPrinter::LineBegin(const string &msg){

    if(numb){
        return;
        }

    if(!operation.empty()){
        LineBreak();
        }

    if(!msg.empty()){
        silent_on_begin = silent;
        operation = msg;
        PrintIndent();
        Print(operation);
        }
    }

//-----------------------------------------------------------------------------
// Complete current status line

void TPrinter::LineEnd(const string &msg){

    if(numb){
        return;
        }

    if(!operation.empty() &&
            (instant_feedback || (!silent && silent_on_begin))){
        PrintIndent();
        Print(operation);
        }

    CompleteTheLine(msg);
    Clear();
    }

//-----------------------------------------------------------------------------
// Add extra line, that will be displayed under the main status line

void TPrinter::LineAddExtra(const string &msg){

    if(numb){
        return;
        }

    if(instant_feedback){
        PrintExtraLine(msg);
        }
    else{
        extra.push_back(msg);
        }
    }

//-----------------------------------------------------------------------------
// Clear vector of extra messages and put the new one in

void TPrinter::LineRewriteExtra(const string &msg){

    if(numb){
        return;
        }

    extra.clear();
    extra.push_back(msg);
    }

//-----------------------------------------------------------------------------
// Break current status line

void TPrinter::LineBreak(void){

    if(numb){
        return;
        }

    Clear();
    Print("\n");
    }

//-----------------------------------------------------------------------------
// Begin a new line with the same message on the left side

void TPrinter::LineRestart(void){

    if(numb){
        return;
        }

    if(!operation.empty()){
        LineBreak();
        }
    LineBegin(operation0);
    }

//-----------------------------------------------------------------------------
// Display additional info right away, not after printing the completion status

void TPrinter::StartInstantFeedback(const string &trivial_msg){

    if(numb){
        return;
        }

    CompleteTheLine(trivial_msg);
    extra.clear();
    instant_feedback = true;
    }

//-----------------------------------------------------------------------------
// Print divider on entire terminal line

void TPrinter::Div(const char s) const{

    if(numb){
        return;
        }

    Print(string(GetTerminalWidth(), s));
    }

//-----------------------------------------------------------------------------
// Clear data fields

void TPrinter::Clear(void){

    operation0 = operation;
    operation.clear();
    extra.clear();
    instant_feedback = false;
    silent_on_begin = false;
    }

//-----------------------------------------------------------------------------
