/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef lcontrolpp_printH
#define lcontrolpp_printH

#include "libcontrol++.h"
#include <vector>

#define DEFAULT_TAB_SIZE  4
#define TERM_MAX_WIDTH    256 // Limit of value returned by the GetTerminalWidth

//-----------------------------------------------------------------------------

// Get terminal width
int GetTerminalWidth(void);

//-----------------------------------------------------------------------------
// Printer of the complex messages

class TPrinter{

    private:
        bool silent = false;            // No output
        bool silent0 = false;           // Prev value of 'silent'
        bool silent_on_begin = false;   // Value of 'silent flag at the moment of operation printing
        bool numb = false;              // Don't do a thing
        bool plain = false;             // Plain output
        bool instant_feedback = false;  // Print additional info right away for
                                        //  thecurrent operation
        char tie_symbol = '.';  // Symbol used to tie left and right parts
                                //  of the status line
        vector<string> indent;  // Indentation stack
        string operation;       // Left part of the status line
        string operation0;      // Prev operation
        vector<string> extra;   // Extra info displayed
                                //  under the main status line

        // Print according to current mode
        void Print(const string &str) const;
        // Print line of extra info
        void PrintExtraLine(const string &ex) const;
        // Print indentation units
        void PrintIndent(void) const;
        // Get a full size of concatenated indentation strings
        size_t IndentSize(void) const;
        // Print left part and right part of the text tied with the 'tie'
        // on entire terminal line
        void CompleteTheLine(const string &status);
        // Form special message
        string SpecMessage(const string &preface, const string& msg) const;
        // Clear data fields
        void Clear(void);

    public:
        // Get current left part of the status line
        string Operation(void) const{return operation;}
        // Get vector of extra lines
        vector<string> Extra(void) const{return extra;}
        // Modifiers
        void Silent(const bool flag){
            silent0 = silent;
            silent = flag;
            }
        void ReSilent(void){
            Silent(silent0);
            }
        void Numb(const bool flag){numb = flag;}
        void Plain(const bool flag){plain = flag;}
        void TieSymbol(const char s){tie_symbol = s;}
        // Access to the mode-flags
        bool IsSilent(void) const{return silent;}
        bool IsPlain(void) const{return plain;}
        // Start a new status line
        void LineBegin(const string &msg);
        // Complete current status line
        void LineEnd(const string &msg="");
        // Add extra line, that will be displayed under the main status line
        void LineAddExtra(const string &msg);
        // Add extra error line
        void LineAddErr(const string &msg=""){
            LineAddExtra(SpecMessage("ERROR", msg));
            }
        void LineAddErr(const vector<string> &msg){
            for(const auto &m : msg){
                LineAddErr(m);
                }
            }
        // Add extra warning line
        void LineAddWarn(const string &msg=""){
            LineAddExtra(SpecMessage("WARNING", msg));
            }
        void LineAddWarn(const vector<string> &msg){
            for(const auto &m : msg){
                LineAddWarn(m);
                }
            }
        // Clear vector of extra messages and put the new one in
        void LineRewriteExtra(const string &msg);
        // Break current status line
        void LineBreak(void);
        // Begin a new line with the same message on the left side
        void LineRestart(void);
        // Raise one level of indentation (and optionally change size of all
        // indents)
        void IndentAdd(const size_t size=DEFAULT_TAB_SIZE){
            indent.push_back(string(size > 0 ? size : DEFAULT_TAB_SIZE, ' '));
            }
        // Drop one level of indentation
        void IndentRem(void){
            if(!indent.empty()){
                indent.pop_back();
                }
            }
        // Clear the stack of indentation strings
        void IndentClear(void){indent.clear();}
        // Display additional info right away, not after printing a status
        void StartInstantFeedback(const string &trivial_msg="[IN PROGRESS]");
        // Print error message
        void Err(const string &msg=""){
            LineEnd(SpecMessage("ERROR", msg));
            }
        // Print warning message
        void Warn(const string &msg=""){
            LineEnd(SpecMessage("WARNING", msg));
            }
        // Print message of seccessfull result
        void Success(const string &msg=""){
            LineEnd(SpecMessage("DONE", msg));
            }
        // Print divider on entire terminal line
        void Div(const char s='=') const;

        TPrinter(){
            cout << unitbuf; // Automaticly flushing the output
            }
        ~TPrinter(){}
    };

//-----------------------------------------------------------------------------

#endif
