/*
    control++ : System configuration tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "libcontrol++.h"
#include "text.h"
#include <algorithm>
#include <regex>

//-----------------------------------------------------------------------------
// Check if specified char is tab or space symbol

bool IsTabOrSpace(char c){

    return c == '\t' || c == ' ';
    };

//-----------------------------------------------------------------------------
// Convert string to uppercase

string ToUpper(string str){

    for(auto &c : str){
        c = toupper(c);
        }

    return str;
    }

//-----------------------------------------------------------------------------
// Convert string to lowercase

string ToLower(string str){

    for(auto &c : str){
        c = tolower(c);
        }

    return str;
    }

//-----------------------------------------------------------------------------
// Mirror single quotes

string MirrorSingleQuotes(const string &line){

    string new_line;
    new_line.reserve(line.size());

    for(const auto c : line){
        if(c == '\''){
            new_line += "'";
            }
        new_line += c;
        }

    return new_line;
    }

//-----------------------------------------------------------------------------
// Get position of n-th ocurrance of a substring in a string

size_t FindSubstrNthOccurrence(const size_t n, const string &subs,
        const string &str){

    if(subs.empty() || n < 1){
        return string::npos;
        }

    size_t pos = 0;
    size_t count = 0;

    while(count < n){
        if(pos >= str.size()){
            return string::npos;
            }
        pos = str.find(subs, pos);
        if(pos == string::npos){
            return string::npos;
            }
        pos++;
        count++;
        }

    return pos - 1;
    }

//-----------------------------------------------------------------------------
// Search for a next word in a given line and return an index of a last
// character of this word

size_t GetWordFromLine(const string &line, string &word, size_t start,
        size_t end){

    word.clear();

    if(line.empty()){
        return string::npos;
        }

    if(end == string::npos){
        end = line.size();
        }

    size_t index = string::npos;

    // For all characters from the start index to the end index
    for(size_t i = start; i < end && i < line.size(); i++){
        const char c = line.at(i);
        // If allowed symbols
        if(!isspace(c)){
            word.push_back(c);
            index = i;
            }
        // Break if word is not empty and there is unallowed symbol
        // or there is a comment
        else if(!word.empty() || c == '#'){
            break;
            }
        }

    return index;
    }

//-----------------------------------------------------------------------------
// Trim string

// Trim chars from the left
void LTrim(string &s, const string &chars){
    Trim(s, chars, true, false);
    }

// Trim chars from the right
void RTrim(string &s, const string &chars){
    Trim(s, chars, false, true);
    }

// Trim chars from both sides
void Trim(string &s, const string &chars, const bool ltrim, const bool rtrim){

    auto not_searched_char = [&](int ch){
        if(chars.empty()){
            return isspace(ch) == 0;
            }
        return chars.find(ch) == string::npos;
        };

    if(ltrim){
        auto it = find_if(s.begin(), s.end(), not_searched_char);
        s.erase(s.begin(), it);
        }

    if(rtrim){
        auto rit = find_if(s.rbegin(), s.rend(), not_searched_char);
        s.erase(rit.base(),s.end());
        }
    }

//-----------------------------------------------------------------------------
// Split given string by extracting substrings placed between occurrences
// of a specified delimiter

vector<string> Split(const string &str, const string &delim){

    vector<string> res;

    if(delim.empty()){
        string s = str;
        smatch m;
        regex e(R"(\s+)");
        for(;;){
            if(!regex_search(s, m, e)){
                res.push_back(s);
                break;
                }
            res.push_back(m.prefix().str());
            s = m.suffix().str();
            }
        }
    else{
        for(size_t i = 0, i0 = 0;;){
            i = str.find(delim, i0);
            if(i == string::npos){
                res.push_back(str.substr(i0));
                break;
                }
            res.push_back(str.substr(i0, i - i0));
            i0 = i + delim.size();
            }
        }

    return res;
    }

//-----------------------------------------------------------------------------
// Cut comment from a given line

string CutComment(const string &line, const string &comment_designation){

    size_t pos = 0;

    for(;;){
        // Searching for a comment symbol
        pos = line.find(comment_designation, pos);
        if(pos == string::npos){
            break;
            }
        if(!Escaped(line, pos)){
            return line.substr(0, pos);
            }
        }

    return line;
    }

//-----------------------------------------------------------------------------
// Check that there are specified symbols in a given string

bool CheckSymbols(const string &str, const set<char> &symbols){

    for(const auto c : symbols){
        if(str.find(c) != string::npos){
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Escape symbols of a given set (control++ special symbols by default)
// in a given string

string Esc(const string &str, const set<char> &esc_symbols){

    string esc_str;

    for(const auto c : str){
        if(esc_symbols.find(c) != esc_symbols.end()){
            esc_str += '\\';
            }
        esc_str += c;
        }

    return esc_str;
    }

//-----------------------------------------------------------------------------
// Count escaping symbols

size_t CountEsc(const string &str, const size_t stop){

    if(stop == 0 || stop >= str.size()){
        return 0;
        }

    size_t counter = 0;

    for(size_t i = (stop - 1); i >= 0; i--){
        if(str.at(i) != '\\'){
            break;
            }
        counter++;
        }

    return counter;
    }

//-----------------------------------------------------------------------------
// Check if symbol is escaped

bool Escaped(const string &str, const size_t index,
        const set<char> &esc_symbols){

    return index < str.size() &&
        esc_symbols.find(str.at(index)) != esc_symbols.end() &&
        CountEsc(str, index) % 2 != 0;
    }

//-----------------------------------------------------------------------------
// Remove an escaping of symbols of a given set (control++ special symbols by
// default) of a given string

string Unesc(const string &str, const set<char> &esc_symbols){

    string unesc_str;

    for(size_t i = 0; i < str.size(); i++){
        if(Escaped(str, i, esc_symbols)){
            // The "unesc_str" string can't be empty at this pont
            unesc_str.back() = str.at(i);
            }
        else{
            unesc_str += str.at(i);
            }
        }

    return unesc_str;
    }

//-----------------------------------------------------------------------------
// Find position of a first unescaped quotation mark in a given string
// starting from the 'pos'

size_t UnescapedQtnMarkPos(const string &str, size_t pos,
        const bool double_quote){

    const char quote = double_quote ? '"' : '\'';

    for(;;){
        if(pos >= str.size()){
            break;
            }
        const auto i = str.find(quote, pos);
        if(i == string::npos || !Escaped(str, i)){
            return i;
            }
        pos = i + 1;
        }

    return string::npos;
    }

//-----------------------------------------------------------------------------
